#include "hex6x12.h"
#include "syscall.h"
#include "validaddr.h"
#include "hexdump.h"
#include "validaddr.h"

/*
 * typedef unsigned long ulong;
 * typedef unsigned short ushort;
 * typedef unsigned char uchar;
 */
 
#define XCHAR_WIDTH	7
#define XCHAR_HEIGHT	12
#define BUF_WIDTH	(XCHAR_WIDTH * 8)
#define BUF_HEIGHT	(XCHAR_HEIGHT * 4)

/* #define ISKANJI1(c)	( ((c) > 0x80 && (c) < 0xa0) || ((c) > 0xdf && (c) < 0xfe) ) */
#define ISKANJI1(c)	( (c) >= 0x81 && (c) <= 0x84 || (c) >= 0x88 && (c) <= 0x98 )

static ushort DmaBuffer[BUF_WIDTH * BUF_HEIGHT];
static enum { BIGENDIAN, LITTLEENDIAN } Endian = BIGENDIAN;


static void XdumpPutNibble(int y, int x, int n, ushort fg, ushort bg)
{
	const uchar *ch = &Hex6x12[9 * n];
	uchar mask = 0x80;
	int h, v;

	for (v = 0 ; v < XCHAR_HEIGHT - 1 ; v ++) {
		for (h = 0 ; h < XCHAR_WIDTH - 1 ; h ++) {
			DmaBuffer[(y + v) * BUF_WIDTH + x + h] =
				 (*ch & mask) ? fg : bg;
			mask >>= 1;
			if (mask == 0) {
				ch ++;
				mask = 0x80;
			}
		}
		DmaBuffer[(y + v) * BUF_WIDTH + x + h] = bg;
	}
	for (h = 0 ; h < XCHAR_WIDTH ; h ++) {
		DmaBuffer[(y + v) * BUF_WIDTH + x + h] = bg;
	}
	return;
}

static void XdumpPutChar(int y, int x, uchar n, ushort fg, ushort bg)
{
	XdumpPutNibble(y, x, (n >> 4) & 0xf, fg, bg);
	XdumpPutNibble(y, x + XCHAR_WIDTH, n & 0xf, fg, bg);
}

static void XdumpPutShort(int y, int x, ushort n, ushort fg, ushort bg)
{
	XdumpPutChar(y, x, n >> 8, fg, bg);
	XdumpPutChar(y, x + XCHAR_WIDTH*2, n & 0xff, fg, bg);
}

static void XdumpPutLong(int y, int x, ulong n, ushort fg, ushort bg)
{
	XdumpPutShort(y, x, n >> 16, fg, bg);
	XdumpPutShort(y, x + XCHAR_WIDTH*4, n & 0xffff, fg, bg);
}

static void XdumpPutPtr(int y, int x, const uchar *ptr, int sz, ushort fg, ushort bg)
{
	while (sz > 0) {
		XdumpPutChar(y, x, *ptr, fg, bg);
		ptr ++;
		x += XCHAR_WIDTH*2;
		sz --;
	}
}

static void ClearDmaBuffer(ushort bg)
{
	int i;
	ushort *ptr = DmaBuffer;
	for (i = BUF_WIDTH * BUF_HEIGHT ; i > 0 ; i --) *ptr ++ = bg;
	return;
}

/* Simple HEX DUMP */
void Xdump(ulong startAddr, ulong dispAddr, int x, int y, ushort fg, ushort bg)
{
	ulong i;
	int block;
	ulong *ptr;

	/* ADDR */
	for (i = 0 ; i < 4 ; i ++) {
		XdumpPutLong(XCHAR_HEIGHT * i, 0, 
			dispAddr + i*16, fg, bg);
	}
	mem2vram(x, y, BUF_WIDTH, BUF_HEIGHT, (long *)DmaBuffer);
	GPU_sync();

	for (block = 0 ; block < 4 ; block ++) {
		ClearDmaBuffer(bg);
		for (i = 0 ; i < 4 ; i ++) {
			ptr = (ulong *)startAddr;
			ptr += block;
			ptr += 4 * i;
			if (IsAddrValid((ulong)ptr)) {
				if (Endian == BIGENDIAN)
					XdumpPutPtr(XCHAR_HEIGHT * i, 0,
						(uchar *)ptr, 4, fg, bg);
				else	XdumpPutLong(XCHAR_HEIGHT * i, 0,
						*ptr, fg, bg);
			}
		}
		mem2vram(x + (BUF_WIDTH + XCHAR_WIDTH) * (1 + block), y,
			BUF_WIDTH, BUF_HEIGHT, (long *)DmaBuffer);
		GPU_sync();
	}

	return;
}

/* RB*/
void XdumpSetup(int x, int y, ushort fg, ushort bg)
{
	int i;
	const uchar *ch;
	uchar mask;
	int h, v;
	ushort *ptr = DmaBuffer;
	GPU_sync();
	for (i = 0 ; i < 4 ; i ++) {
		ch = &Hex6x12[9 * 16];
		mask = 0x80;
		for (v = 0 ; v < XCHAR_HEIGHT - 1 ; v ++) {
			for (h = 0 ; h < XCHAR_WIDTH - 1 ; h ++) {
				*ptr ++ = (*ch & mask) ? fg : bg;
				mask >>= 1;
				if (mask == 0) {
					ch ++;
					mask = 0x80;
				}
			}
		}
		for (h = 0 ; h < XCHAR_WIDTH - 1 ; h ++) *ptr ++ = bg;
	}
	mem2vram(x + XCHAR_WIDTH * 8, y, 
		XCHAR_WIDTH -1, XCHAR_HEIGHT * 4, (long *)DmaBuffer);
	GPU_sync();

	return;
}


/*
 *	buffer  ptr Ŏw肳ꂽAhX 32bytes ɑ΂
 *	\\ȕɕϊB
 *	1byte ڂQoCgځA܂͍ŏIoCgPoCgڂ
 *	ꍇ\̂ōvPWoCg̕쐬B
 *	kanji == 0 ̂Ƃ ANK ̂ݕ\B
 */
void MakePrintable(uchar *buffer, ulong ptr, int kanji)
{
#define TARG_BYTES	32
	uchar *p = (uchar *)ptr;
	int i;
	
	buffer[0] = buffer[TARG_BYTES + 1] = ' ';
	buffer[TARG_BYTES + 2] = '\0';
	
	if (kanji == 0) {	/* ANK ̂ */
		for (i = 0 ; i < TARG_BYTES ; i ++) {
			if (IsAddrValid(ptr + i)) {
				buffer[i + 1] = isprint(p[i]) ? p[i] : '.';
			} else buffer[i + 1] = ' ';
		}
		return;
	}
	/* ܂ */

	i = 0;
	buffer[0] = ' ';
	if (IsAddrValid(ptr - 1) && ISKANJI1(p[-1])) {
		/* PoCgOPoCgڂȂB */
		if (IsReallyKanji1(ptr - 1)) {
			buffer[0] = p[-1];
			buffer[1] = p[0];
			i = 1;
		}
	}
	while (i < TARG_BYTES) {
		if (IsAddrValid(ptr + i) == 0) {
			buffer[i + 1] = ' ';
			i ++; continue;
		}
		if (ISKANJI1(p[i]) && IsAddrValid(ptr + i + 1)) {
			/*  */
			buffer[i + 1] = p[i]; i ++;
			buffer[i + 1] = p[i]; i ++;
		} else {
			buffer[i + 1] = isprint(p[i]) ? p[i] : '.';
			i ++;
		}
	}
	return;
}

int IsReallyKanji1(ulong p)
{
	unsigned char *ptr = (unsigned char *)p;
	
	if (IsAddrValid(p + 1) == 0) return 0;
	if (IsAddrValid(p) == 0) return 0;
	while (ISKANJI1(*ptr)) {
		ptr -- ;
		if (IsAddrValid((ulong)ptr) == 0) break;
	}
	return (p - (ulong)ptr) & 1;
}

int isprint(int c)
{
	if (c & 0x80) {
		c >>= 5;
		if (c == 4 || c == 7) return 0;
		return 1;	/* 5,6 = pJi */
	}
	if (c < 0x20) return 0;
	if (c == 0x7f) return 0;
	
	return 1;
}



